#!/usr/bin/python3
# -*- coding: utf-8 -*-

import gi
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
gi.require_version('GimpUi', '3.0')
from gi.repository import GimpUi
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gegl
import sys
import time

plug_in_proc   = "plug-in-joey-py3-sprites-to-layers"
plug_in_binary = "py3-sprites-to-layers"

def sprites_to_layers(procedure, run_mode, image, drawables, config, data):
    if run_mode == Gimp.RunMode.INTERACTIVE:
        GimpUi.init(plug_in_binary)
        dialog = GimpUi.ProcedureDialog.new(procedure, config, "Sprites to Layers")
        
        dialog.fill(['sprites-group-layer-name'
                    , 'sprites-prefix-name'
                    , 'small-sprites-group-layer-name'
                    , 'small-sprites-prefix-name'
                    , 'sample'
                    , 'minimal-dimensions'])
        if not dialog.run():
            dialog.destroy()
            return procedure.new_return_values(Gimp.PDBStatusType.CANCEL, None)
        else:
            dialog.destroy()
    
    # Get dialog parameters.
    sprites_group_layer_name       = config.get_property('sprites-group-layer-name')
    sprites_prefix_name            = config.get_property('sprites-prefix-name')
    small_sprites_group_layer_name = config.get_property('small-sprites-group-layer-name')
    small_sprites_prefix_name      = config.get_property('small-sprites-prefix-name')
    sample                         = config.get_property('sample')
    minimal_dimensions             = config.get_property('minimal-dimensions')
    
    image.undo_group_start()
    
    # Start timer.
    time_start = time.time()
    
    layer = image.get_selected_layers()[0]
    # Check image type (RGB/Greyscale or Indexed) and if layer has alpha, if not stop this script with a warning.
    no_alpha_warning='Warning\nThis layer has no alpha, please add an alpha channel for this script to work.'
    image_base_type = image.get_base_type()
    match image_base_type:
        case 0:
            image_mode = 'RGB'
            channels = 4
            if layer.get_bpp() != 4:
                Gimp.message(no_alpha_warning)
                return
        case 1:
            image_mode = 'Greyscale'
            channels = 2
            if layer.get_bpp() != 2:
                Gimp.message(no_alpha_warning)
                return
        case 2:
            image_mode = 'Indexed'
            channels = 2
            if layer.get_bpp() != 2:
                Gimp.message(no_alpha_warning)
                return
    
    layer_copy = layer.copy()
    image.insert_layer(layer_copy, layer.get_parent(), 0)
    
    # Store Context getters for pdb.gimp_image_select_contiguous_color(), to restore them after operation.
    previous_antialias           = Gimp.context_get_antialias()
    previous_feather             = Gimp.context_get_feather()
    previous_sample_merged       = Gimp.context_get_sample_merged()
    previous_sample_criterion    = Gimp.context_get_sample_criterion()
    previous_sample_threshold    = Gimp.context_get_sample_threshold()
    previous_sample_transparent  = Gimp.context_get_sample_transparent()
    
    # Context setters for pdb.gimp_image_select_contiguous_color().
    Gimp.context_set_antialias(0)
    Gimp.context_set_feather(0)
    Gimp.context_set_sample_merged(0)
    Gimp.context_set_sample_criterion(0)
    Gimp.context_set_sample_threshold(1)
    Gimp.context_set_sample_transparent(0)
    
    # Counters for objects in layer groups.
    counter_good = 1
    counter_bad = 1
    
    layer_width = layer_copy.get_width()
    layer_height = layer_copy.get_height()
    
    # Loop over pixels.
    for pixel_y in range(0, layer_height, sample):
        buffer = layer_copy.get_buffer()
        row = buffer.get(Gegl.Rectangle.new(x=0, y=pixel_y, width=layer_width, height=1), 1, None, Gegl.AbyssPolicy.NONE)
        for pixel_alpha in range((channels - 1), len(row), channels * sample):
            # If pixel is not completely transparent select it and neighbouring non-transparent pixels.
            if row[pixel_alpha] > 0:
                pixel_x = (pixel_alpha - (channels - 1))/(channels)
                image.select_contiguous_color(2, layer_copy, pixel_x, pixel_y)
                # Create a layer for an object and assign it to 'small objects' layer group or
                # 'objects layer group' based on criteria "min_dimensions".
                object_layer = layer_copy.copy()
                _, x1, y1, x2, y2 = layer_copy.mask_bounds()
                
                # 'Small objects' layer group.
                if x2 - x1 < minimal_dimensions and y2 - y1 < minimal_dimensions:
                    if image_mode == 'RGB' or image_mode == 'Greyscale':
                        # Create layer_group_bad.
                        if counter_bad == 1:
                            #layer_group_bad = Gimp.GroupLayer.new(image, 'small objects')
                            layer_group_bad = Gimp.GroupLayer.new(image, small_sprites_group_layer_name)
                            image.insert_layer(layer_group_bad, layer.get_parent(), 0)
                        # Add object to layer_group_bad at the last position.
                        image.insert_layer(object_layer, layer_group_bad, len(layer_group_bad.get_children()))
                    # Indexed mode does not allow GroupLayers, so object layer added to image.
                    if image_mode == 'Indexed':
                        image.insert_layer(object_layer, layer.get_parent(), 0)
                    # Name object/sprite "trash (position_x, position_y, size_x, size_y)".
                    object_layer.set_name("{} {:04d} ({:d}, {:d}, {:d}, {:d})".format(small_sprites_prefix_name, counter_good, x1, y1, x2 - x1, y2 - y1))
                    counter_bad += 1
                
                # 'Objects' layer group.
                else:
                    if image_mode == 'RGB' or image_mode == 'Greyscale':
                        # Create layer_group_good.
                        if counter_good == 1:
                            #layer_group_good = Gimp.GroupLayer.new(image, 'objects')
                            layer_group_good = Gimp.GroupLayer.new(image, sprites_group_layer_name)
                            image.insert_layer(layer_group_good, layer.get_parent(), 0)
                        # Add object to layer_group_good at the last position.
                        image.insert_layer(object_layer, layer_group_good, len(layer_group_good.get_children()))
                    # Indexed mode does not allow GroupLayers, so object layer added to image.
                    if image_mode == 'Indexed':
                        image.insert_layer(object_layer, layer.get_parent(), 0)
                    # Name object/sprite "object (position_x, position_y, size_x, size_y)".
                    object_layer.set_name("{} {:04d} ({:d}, {:d}, {:d}, {:d})".format(sprites_prefix_name, counter_good, x1, y1, x2 - x1, y2 - y1))
                    counter_good += 1
                
                # "Auto crop" object layer.
                object_layer.resize(x2 - x1, y2 - y1, -x1, -y1)
                # Remove/erase selection from layer_copy
                layer_copy.edit_clear()
                # Get a refreshed buffer.
                buffer = layer_copy.get_buffer()
                row = buffer.get(Gegl.Rectangle.new(x=0, y=pixel_y, width=layer_width, height=1), 1, None, Gegl.AbyssPolicy.NONE)
    
    # Remove layer_copy from canvas and memory, delete buffer and row.
    image.remove_layer(layer_copy)
    del(layer_copy)
    del(buffer)
    del(row)
    
    # Restore context setters to before operation.
    Gimp.context_set_antialias(previous_antialias)
    Gimp.context_set_feather(previous_feather)
    Gimp.context_set_sample_merged(previous_sample_merged)
    Gimp.context_set_sample_criterion(previous_sample_criterion)
    Gimp.context_set_sample_threshold(previous_sample_threshold)
    Gimp.context_set_sample_transparent(previous_sample_transparent)
    
    # End timer and display number of objects, small objects and timer seconds up to milliseconds in console.
    time_end = time.time()
    Gimp.message('INFO:\n-{:d} objects found\n'.format(counter_good - 1) +
                 '-{:d} small objects found\n'.format(counter_bad - 1) +
                 '-time taken: {:.3f} seconds'.format(time_end - time_start))
    
    image.undo_group_end()
    return procedure.new_return_values(Gimp.PDBStatusType.SUCCESS, None)

class SpritesToLayers (Gimp.PlugIn):
    def do_query_procedures(self):
        return [ plug_in_proc ]
    
    def do_create_procedure(self, name):
        procedure = None
        if name == plug_in_proc:
            procedure = Gimp.ImageProcedure.new(self, name,
                                                Gimp.PDBProcType.PLUGIN,
                                                sprites_to_layers, None)
            procedure.set_sensitivity_mask (Gimp.ProcedureSensitivityMask.DRAWABLE |
                                            Gimp.ProcedureSensitivityMask.NO_DRAWABLES)
            procedure.set_attribution("joey", "joey, joey project", "2025")
            procedure.set_menu_label("Sprites to layers")
            procedure.add_menu_path ("<Image>/Layer")
            procedure.set_documentation ("Extract sprites to layers",
                                            "Separates objects or sprites from one layer to their own layer. Handy for Spritesheets.",
                                            None)
            # dialog box parameters
            procedure.add_string_argument  (  "sprites-group-layer-name"
                                            , "Sprites GroupLayer name"
                                            , "Name of sprites group layer."
                                            , "Sprites"
                                            , GObject.ParamFlags.READWRITE)
            procedure.add_string_argument  (  "small-sprites-group-layer-name"
                                            , "Small Sprites GroupLayer name"
                                            , "Name of small sprites group layer."
                                            , "Small Sprites"
                                            , GObject.ParamFlags.READWRITE)
            procedure.add_string_argument  (  "sprites-prefix-name"
                                            , "Sprite Prefix"
                                            , "Prefix of extracted sprites."
                                            , "Sprite"
                                            , GObject.ParamFlags.READWRITE)
            procedure.add_string_argument  (  "small-sprites-prefix-name"
                                            , "Small Sprite Prefix"
                                            , "Prefix of extracted small sprites."
                                            , "Small Sprite"
                                            , GObject.ParamFlags.READWRITE)
            procedure.add_int_argument     (  "sample", "Sample every nth pixel"
                                            , "This option will sample every nth pixel, higher numbers will speed up the script, but too high values will risk skipping sprites, best to keep this lower than the Minimal Dimensions."
                                            , 1, 1000, 2, GObject.ParamFlags.READWRITE)
            procedure.add_int_argument     (  "minimal-dimensions"
                                            , "Minimal Dimensions"
                                            , "Minimal square dimensions to be considered a good sprite, for example 20x20 pixels."
                                            , 1, 1000, 20, GObject.ParamFlags.READWRITE)
        return procedure

Gimp.main(SpritesToLayers.__gtype__, sys.argv)